import datetime as dt
import os

from pathlib import Path

import pandas as pd
import pytest

from lumen.sources.base import Source
from lumen.state import state
from lumen.transforms.sql import SQLLimit


@pytest.fixture
def source(make_filesource):
    root = os.path.dirname(__file__)
    return make_filesource(root)


@pytest.fixture
def expected_df(column_value_type):
    df = pd._testing.makeMixedDataFrame()
    column, value, type = column_value_type

    if type == 'single_value':
        return df[df[column] == value]

    elif type == 'range':
        begin, end = value
        return df[(df[column] >= begin) & (df[column] <= end)]

    elif type == 'range_list':
        conditions = False
        for range in value:
            begin, end = range
            conditions |= ((df[column] >= begin) & (df[column] <= end))
        return df[conditions]

    elif type == 'list':
        return df[df[column].isin(value)]

    elif type == 'date':
        return df[df[column] == pd.to_datetime(value)]

    elif type == 'date_range':
        begin, end = value
        return df[(df[column] >= pd.to_datetime(begin)) & (df[column] <= pd.to_datetime(end))]

    return df


def test_source_resolve_module_type():
    assert Source._get_type('lumen.sources.base.Source') is Source


@pytest.mark.parametrize("filter_col_A", [314, (13, 314), ['A', 314, 'def']])
@pytest.mark.parametrize("filter_col_B", [(3, 15.9)])
@pytest.mark.parametrize("filter_col_C", [[1, 'A', 'def']])
@pytest.mark.parametrize("sql_transforms", [(None, None), (SQLLimit(limit=100), SQLLimit(limit=100))])
def test_file_source_table_cache_key(source, filter_col_A, filter_col_B, filter_col_C, sql_transforms):
    t1, t2 = sql_transforms
    kwargs1 = {}
    kwargs2 = {}
    if t1 is not None:
        kwargs1['sql_transforms'] = [t1]
    if t2 is not None:
        kwargs2['sql_transforms'] = [t2]
    key1 = source._get_key('test', A=filter_col_A, B=filter_col_B, C=filter_col_C, **kwargs1)
    key2 = source._get_key('test', A=filter_col_A, B=filter_col_B, C=filter_col_C, **kwargs2)
    assert key1 == key2


@pytest.mark.parametrize(
    "column_value_type", [
        ('A', 1, 'single_value'),
        ('A', (1, 3), 'range'),
        ('A', (1, 2), 'range'),
        ('A', [(0, 1), (3, 4)], 'range_list'),
        ('C', 'foo2', 'single_value'),
        ('C', ['foo1', 'foo3'], 'list'),
        ('D', dt.datetime(2009, 1, 2), 'single_value'),
        ('D', (dt.datetime(2009, 1, 2), dt.datetime(2009, 1, 5)), 'range'),
        ('D', [dt.datetime(2009, 1, 2), dt.datetime(2009, 1, 5)], 'list'),
        ('D', dt.datetime(2009, 1, 2), 'date'),
        ('D', (dt.date(2009, 1, 2), dt.date(2009, 1, 5)), 'date_range'),
    ]
)
@pytest.mark.parametrize("dask", [True, False])
def test_file_source_filter(source, column_value_type, dask, expected_df):
    column, value, _ = column_value_type
    kwargs = {column: value, '__dask': dask}
    filtered = source.get('test', **kwargs)
    pd.testing.assert_frame_equal(filtered, expected_df)


@pytest.mark.parametrize(
    "column_value_type", [
        ('A', 1, 'single_value'),
        ('A', (1, 3), 'range'),
        ('A', (1, 2), 'range'),
        ('A', [(0, 1), (3, 4)], 'range_list'),
        ('C', 'foo2', 'single_value'),
        ('C', ['foo1', 'foo3'], 'list'),
        ('D', dt.datetime(2009, 1, 2), 'single_value'),
        ('D', (dt.datetime(2009, 1, 2), dt.datetime(2009, 1, 5)), 'range'),
        ('D', [dt.datetime(2009, 1, 2), dt.datetime(2009, 1, 5)], 'list'),
        ('D', dt.datetime(2009, 1, 2), 'date'),
        ('D', (dt.date(2009, 1, 2), dt.date(2009, 1, 5)), 'date_range'),
    ]
)
@pytest.mark.parametrize("dask", [True, False])
def test_file_source_get_query_cache(source, column_value_type, dask, expected_df):
    column, value, _ = column_value_type
    kwargs = {column: value}
    source.get('test', __dask=dask, **kwargs)
    cache_key = source._get_key('test', **kwargs)
    assert cache_key in source._cache
    cached_df = source._cache[cache_key]
    if dask:
        cached_df = cached_df.compute()
    pd.testing.assert_frame_equal(cached_df, expected_df)
    cache_key = source._get_key('test', **kwargs)
    assert cache_key in source._cache


@pytest.mark.parametrize(
    "column_value_type", [
        ('A', 1, 'single_value'),
        ('A', [(0, 1), (3, 4)], 'range_list'),
        ('C', ['foo1', 'foo3'], 'list'),
        ('D', (dt.datetime(2009, 1, 2), dt.datetime(2009, 1, 5)), 'range'),
        ('D', dt.datetime(2009, 1, 2), 'date'),
    ]
)
@pytest.mark.parametrize("dask", [True, False])
def test_file_source_clear_cache(source, column_value_type, dask):
    column, value, _ = column_value_type
    kwargs = {column: value}
    source.get('test', __dask=dask, **kwargs)
    cache_key = source._get_key('test', **kwargs)
    assert cache_key in source._cache
    source.clear_cache()
    assert len(source._cache) == 0


def test_file_source_get_query_cache_to_file(make_filesource, cachedir):
    root = os.path.dirname(__file__)
    source = make_filesource(root, cache_dir=cachedir)
    source.get('test', A=(1, 2))

    cache_key = source._get_key('test', A=(1, 2))
    cache_path = Path(cachedir) / f'{cache_key}_test.parq'
    df = pd.read_parquet(cache_path, engine='fastparquet')

    # Patch index names due to https://github.com/dask/fastparquet/issues/732
    df.index.names = [None]
    pd.testing.assert_frame_equal(
        df,
        pd._testing.makeMixedDataFrame().iloc[1:3]
    )


def test_file_source_get_tables(source):
    tables = source.get_tables()
    assert tables == ['test']


def test_file_source_variable(make_variable_filesource):
    root = os.path.dirname(__file__)
    source = make_variable_filesource(root)
    state.variables.tables = {'test': 'test2.csv'}
    df = source.get('test')
    expected = pd._testing.makeMixedDataFrame().iloc[::-1].reset_index(drop=True)
    pd.testing.assert_frame_equal(df, expected)


def test_extension_of_comlicated_url(source):
    url = "https://api.tfl.gov.uk/Occupancy/BikePoints/@{stations.stations.id}?app_key=random_numbers"
    source.tables["test"] = url
    assert source._named_files["test"][1] is None



import datetime as dt
import urllib.parse as urlparse

from collections import defaultdict
from concurrent import futures

import pandas as pd
import panel as pn
import param  # type: ignore
import requests

from ..util import parse_timedelta
from .base import Source, cached


class PrometheusSource(Source):
    """
    `PrometheusSource` allows querying Prometheus PromQL endpoints.

    The `PrometheusSource` is configured to return timeseries about CPU,
    memory and network usage as well as restarts for a list of
    Kubernetes pods specified by ID.
    """

    ae5_source = param.Parameter(doc="""
      An AE5Source instance to use for querying.""")

    cache_per_query = param.Boolean(default=False, doc="""
        Whether to query the whole dataset or individual queries.""")

    ids = param.List(default=[], doc="""
      List of pod IDs to query.""")

    metrics = param.List(
        default=['memory_usage', 'cpu_usage', 'restarts',
                 'network_receive_bytes', 'network_transmit_bytes'],
        doc="Names of metric queries to execute")

    promql_api = param.String(doc="""
      Name of the AE5 deployment exposing the Prometheus API""")

    period = param.String(default='3h', doc="""
      Period to query over specified as a string. Supports:

        - Week:   '1w'
        - Day:    '1d'
        - Hour:   '1h'
        - Minute: '1m'
        - Second: '1s'
    """)

    samples = param.Integer(default=200, doc="""
      Number of samples in the selected period to query. May be
      overridden by explicit step value.""")

    shared = param.Boolean(default=False, readonly=True, doc="""
      PrometheusSource cannot be shared because it has per-user state.""")

    step = param.String(doc="""
        Step value to use in PromQL query_range query.""")

    source_type = 'prometheus'

    _memory_usage_query = """sum by(container_name)
    (container_memory_usage_bytes{job="kubelet",
    cluster="", namespace="default", pod_name=POD_NAME,
    container_name=~"app", container_name!="POD"})"""

    _network_receive_bytes_query = """sort_desc(sum by (pod_name)
    (rate(container_network_receive_bytes_total{job="kubelet", cluster="",
    namespace="default", pod_name=POD_NAME}[1m])))"""

    _network_transmit_bytes_query = """sort_desc(sum by (pod_name)
    (rate(container_network_transmit_bytes_total{job="kubelet", cluster="",
    namespace="default", pod_name=POD_NAME}[1m])))"""

    _cpu_usage_query = """sum by (container_name)
    (rate(container_cpu_usage_seconds_total{job="kubelet", cluster="",
     namespace="default", image!="", pod_name=POD_NAME,
     container_name=~"app", container_name!="POD"}[1m]))"""

    _restarts_query = """max by (container)
     (kube_pod_container_status_restarts_total{job="kube-state-metrics",
    cluster="", namespace="default", pod=POD_NAME,
    container=~"app"})"""

    _metrics = {
        'memory_usage': {
            'query': _memory_usage_query,
            'schema': {"type": "number"}
        },
        'network_receive_bytes': {
            'query': _network_receive_bytes_query,
            'schema': {"type": "number"}
        },
        'network_transmit_bytes': {
            'query': _network_transmit_bytes_query,
            'schema': {"type": "number"}
        },
        'cpu_usage': {
            'query': _cpu_usage_query,
            'schema': {"type": "number"}
        },
        'restarts': {
            'query': _restarts_query,
            'schema': {"type": "number"}
        }
    }

    @property
    def panel(self):
        return pn.Param(self.param, parameters=['period', 'samples', 'step'],
                        sizing_mode='stretch_width', show_name=False)

    def _format_timestamps(self):
        end = dt.datetime.utcnow()
        end_formatted = end.isoformat("T") + "Z"
        period = parse_timedelta(self.period)
        if period is None:
            raise ValueError(f"Could not parse period '{self.period}'. "
                             "Must specify weeks ('1w'), days ('1d'), "
                             "hours ('1h'), minutes ('1m'), or "
                             "seconds ('1s').")
        start = end - period
        start_formatted = start.isoformat("T") + "Z"
        return start_formatted, end_formatted

    def _step_value(self):
        if self.step:
            return self.step
        return int((parse_timedelta(self.period)/self.samples).total_seconds())

    def _url_query_parameters(self, pod_id, query):
        """
        Uses regular expression to map ae5-tools pod_id to full id.
        """
        start_timestamp, end_timestamp = self._format_timestamps()
        regexp = f'anaconda-app-{pod_id}-.*'
        query = query.replace("pod_name=POD_NAME", f"pod_name=~'{regexp}'")
        query = query.replace("pod=POD_NAME", f"pod=~'{regexp}'")
        query_params = {
            'query': query, 'start': start_timestamp,
            'end': end_timestamp, 'step': self._step_value()
        }
        return urlparse.urlencode(query_params)

    def _get_query_url(self, metric, pod_id):
        "Return the full query URL"
        query_template = self._metrics[metric]['query']
        query_params = self._url_query_parameters(pod_id, query_template)
        if self.ae5_source:
            ae5 = self.ae5_source._session
            return f'https://{ae5._k8s_endpoint}.{ae5.hostname}/promql/query_range?{query_params}'
        else:
            return f'{self.promql_api}/query_range?{query_params}'

    def _get_query_json(self, query_url):
        "Function called in parallel to fetch JSON in ThreadPoolExecutor"
        if self.ae5_source:
            response = self.ae5_source._session.session.get(query_url, verify=False)
        else:
            response = requests.get(query_url, verify=False)
        data = response.json()
        if len(data) == 0:
            return None
        return data

    def _json_to_df(self, metric, values):
        "Convert JSON response to pandas DataFrame"
        df = pd.DataFrame(values, columns=['timestamp', metric])
        df[metric] = df[metric].astype(float)
        df['timestamp'] = pd.to_datetime(df['timestamp'], unit='s')
        return df.set_index('timestamp')

    def _fetch_data(self, pod_ids):
        "Returns fetched JSON in dictionary indexed by pod_id then metric name"
        if not pod_ids:
            return {}
        # ToDo: Remove need to slice
        triples = [
            (pod_id, metric, self._get_query_url(metric, pod_id[3:]))
            for pod_id in pod_ids for metric in self.metrics
        ]
        fetched_json = defaultdict(dict)
        with futures.ThreadPoolExecutor(len(triples)) as executor:
            tasks = {
                executor.submit(self._get_query_json, t[2]): t
                for t in triples
            }
            for future in futures.as_completed(tasks):
                (pod_id, metric, query_url) = tasks[future]
                try:
                    fetched_json[pod_id][metric] = future.result()
                except Exception as e:
                    fetched_json[pod_id][metric] = []
                    self.param.warning(
                        f"Could not fetch {metric} for pod {pod_id}. "
                        f"Query used: {query_url}, errored with {type(e)}({e})."
                    )
        return fetched_json

    def _make_query(self):
        json_data = self._fetch_data(self.ids)
        dfs = []
        for pod_id, pod_data in json_data.items():
            df = None
            for metric in self.metrics:
                data_df = self._json_to_df(metric, pod_data[metric])
                if df is None:
                    df = data_df
                else:
                    df = pd.merge(df, data_df, on='timestamp', how='outer')
            df = df.reset_index()
            df.insert(0, 'id', pod_id)
            dfs.append(df)
        if dfs:
            return pd.concat(dfs)
        else:
            return pd.DataFrame(columns=list(self.get_schema('timeseries')))

    def get_schema(self, table=None):
        dt_start, dt_end = self._format_timestamps()
        schema = {
            "id": {
                "type": "string",
                "enum": self.ids
            },
            "timestamp" : {
                "type": "string",
                "format": "date-time",
                "formatMinimum": dt_start,
                "formatMaximum": dt_end
            }
        }
        for k, mdef in self._metrics.items():
            schema[k] = mdef['schema']
        return {"timeseries": schema} if table is None else schema

    @cached
    def get(self, table, **query):
        if table not in ('timeseries',):
            raise ValueError(f"PrometheusSource has no '{table}' table, "
                             "it currently only has a 'timeseries' table.")
        return self._make_query()

    def get_tables(self):
        return list(self._metrics)



import re

import pytest

from lumen.validation import ValidationError


def test_only_bold_whole_words():
    spec = {'title': 'Table', 'source': 'penguin', 'views': [{'type': 'table', 'table': 'penguins'}]}
    msg = 'test\n\n    title: Table\n    source: \x1b[1mpenguin\x1b[0m\n    views:\n    - type: table\n      table: penguins\n'
    with pytest.raises(ValidationError, match=re.escape(msg)):
        raise ValidationError("test", spec, "penguin")



from __future__ import annotations

from typing import (
    TYPE_CHECKING, Any, Dict, List, Tuple, cast,
)
from weakref import WeakKeyDictionary

import panel as pn

from .util import extract_refs, is_ref

if TYPE_CHECKING:
    from bokeh.document import Document  # type: ignore

    from .dashboard import Config
    from .filters.base import Filter
    from .pipeline import Pipeline
    from .sources.base import Source
    from .variables.base import Variables


class _session_state:
    """
    The session_state object holds both global and session specific
    state making it easy to manage access to Sources and Filters
    across components.
    """

    global_sources: Dict[str, Source] = {}

    global_filters: Dict[str, Dict[str, Tuple[Dict[str, Any], Dict[str, Any]]]] = {}

    _configs: WeakKeyDictionary[Document, Config] = WeakKeyDictionary()
    _config: Config | None = None

    _specs: WeakKeyDictionary[Document, Dict[str, Any]] = WeakKeyDictionary()
    _spec: Dict[str, Any] = {}

    _loadings: WeakKeyDictionary[Document, Dict[str, pn.pane.HTML]] = WeakKeyDictionary()
    _loading: pn.pane.HTML | None = None

    _sources: WeakKeyDictionary[Document, Dict[str, Source]] = WeakKeyDictionary()
    _source: Dict[str, Source] = {}

    _pipelines: WeakKeyDictionary[Document, Dict[str, Pipeline]] = WeakKeyDictionary()
    _pipeline: Dict[str, Pipeline] = {}

    _filters: WeakKeyDictionary[Document, Dict[str, Dict[str, Filter]]] = WeakKeyDictionary()
    _filter: Dict[str, Dict[str, Filter]] = {}

    _variables: WeakKeyDictionary[Document, Variables] = WeakKeyDictionary()
    _variable: Variables | None = None

    @property
    def config(self) -> Config | None:
        if pn.state.curdoc is None:
            return self._config
        return self._configs.get(pn.state.curdoc, self._config)

    @config.setter
    def config(self, config: Config):
        if pn.state.curdoc is None:
            self._config = config
        else:
            self._configs[pn.state.curdoc] = config

    @property
    def filters(self) -> Dict[str, Dict[str, Filter]]:
        if pn.state.curdoc is None:
            if self._filter is None:
                self._filter = self._global_filters
            return self._filter
        elif pn.state.curdoc not in self._filters:
            self._filters[pn.state.curdoc] = self._global_filters
        return self._filters[pn.state.curdoc]

    @property
    def loading_msg(self):
        if pn.state.curdoc is None:
            return self._loading
        return self._loadings.get(pn.state.curdoc)

    @loading_msg.setter
    def loading_msg(self, loading_msg: pn.pane.HTML):
        if pn.state.curdoc is None:
            self._loading = loading_msg
        else:
            self._loadings[pn.state.curdoc] = loading_msg

    @property
    def pipelines(self) -> Dict[str, Pipeline]:
        if pn.state.curdoc is None:
            return self._pipeline
        elif pn.state.curdoc not in self._pipelines:
            self._pipelines[pn.state.curdoc] = dict(self._pipeline)
        return self._pipelines[pn.state.curdoc]

    @property
    def sources(self) -> Dict[str, Source]:
        if pn.state.curdoc is None:
            return self._source
        elif pn.state.curdoc not in self._sources:
            self._sources[pn.state.curdoc] = dict(
                self.global_sources, **self._source
            )
        return self._sources[pn.state.curdoc]

    @property
    def spec(self) -> Dict[str, Any]:
        if pn.state.curdoc is None:
            return self._spec
        return self._specs.get(pn.state.curdoc, self._spec)

    @spec.setter
    def spec(self, spec: Dict[str, Any]):
        if pn.state.curdoc is None:
            self._spec = spec
        else:
            self._specs[pn.state.curdoc] = spec

    @property
    def variables(self) -> Variables:
        from .variables import Variables
        if self._variable is None:
            self._variable = Variables.create_variables()
        if pn.state.curdoc is None:
            return self._variable  # type: ignore
        elif pn.state.curdoc not in self._variables:
            self._variables[pn.state.curdoc] = variables = Variables.create_variables()
            for var in self._variable._vars.values():  # type: ignore
                variables.add_variable(var)
        return self._variables[pn.state.curdoc]

    @property
    def _global_filters(self):
        from .filters import Filter
        return {
            source: {
                name: Filter.from_spec(spec, schema)
                for name, (spec, schema) in filters.items()
            }
            for source, filters in self.global_filters.items()
        }

    def to_spec(
        self, auth=None, config=None, defaults=None,
        pipelines={}, sources={}, layouts=[], variables=None
    ):
        """
        Exports the full specification of the supplied components including
        the variable definitions.

        Parameters
        ----------
        auth: lumen.dashboard.Auth
        config: lumen.dashboard.Config
        defaults: lumen.dashboard.Defaults
        variables: lumen.variables.Variables
        pipelines: Dict[str, Pipeline]
        sources: Dict[str, Source]
        layouts: list[Layout]

        Returns
        -------
        Declarative specification of all the supplied components.
        """
        variables = variables or self.variables
        context = {}
        if auth:
            from .dashboard import Auth
            if isinstance(auth, dict):
                context['auth'] = Auth.validate(auth)
            else:
                context['auth'] = auth.to_spec()
        if config:
            from .dashboard import Config
            if isinstance(config, dict):
                context['config'] = Config.validate(config)
            else:
                context['config'] = config.to_spec()
        if defaults:
            from .dashboard import Defaults
            if isinstance(defaults, dict):
                context['defaults'] = Defaults.validate(defaults)
            else:
                context['defaults'] = defaults.to_spec()
        if self.variables._vars:
            context['variables'] = {
                k: v.to_spec() for k, v in self.variables._vars.items()
            }
        if sources:
            context['sources'] = {}
        for k, source in sources.items():
            context['sources'][k] = source.to_spec(context=context)
        if pipelines:
            context['pipelines'] = {}
        for k, pipeline in pipelines.items():
            context['pipelines'][k] = pipeline.to_spec(context=context)
        if layouts:
            context['layouts'] = [
                layout.to_spec(context) for layout in layouts
            ]
        return context

    @property
    def global_refs(self) -> List[str]:
        layout_refs = []
        source_specs = self.spec.get('sources', {})
        for layout in self.spec.get('layouts', []):
            if isinstance(layout.get('source'), str):
                src_ref = layout['source']
                layout = dict(layout, source=source_specs.get(src_ref))
            layout_refs.append(set(extract_refs(layout, 'variables')))

        var_refs = extract_refs(self.spec.get('variables'), 'variables')
        if layout_refs:
            combined_refs = set.union(*layout_refs)
            merged_refs = set.intersection(*layout_refs)
        else:
            combined_refs, merged_refs = set(), set()
        return list(merged_refs | (set(var_refs) - combined_refs))

    def load_global_sources(self, clear_cache: bool = True):
        """
        Loads global sources shared across all layouts.
        """
        from .sources.base import Source
        for name, source_spec in self.spec.get('sources', {}).items():
            if not source_spec.get('shared'):
                continue
            source_spec = dict(source_spec)
            filter_specs: Dict[str, Any] = source_spec.pop('filters', {})
            source = self.global_sources.get(name)
            source_type = Source._get_type(source_spec['type'])
            reinitialize = False
            if source is not None:
                # Check if old source is the same as the new source
                params = {
                    k: v for k, v in source.param.get_param_values()
                    if k != 'name'
                }
                new_spec = dict(source_spec, **{
                    k: v for k, v in source_type.param.get_param_values()
                    if k != 'name'
                })
                reinitialize = params == new_spec
            if self.loading_msg is not None:
                self.loading_msg.object = f'Loading {name} source...'
            if reinitialize or source is None:
                source_spec['name'] = name
                self.global_sources[name] = source = Source.from_spec(source_spec)
            if source.cache_dir and clear_cache:
                source.clear_cache()
            if not filter_specs:
                continue
            tables = set(filter_spec.get('table') for filter_spec in filter_specs.values())
            if None in tables:
                schemas = source.get_schema()
            else:
                schemas = {table: source.get_schema(table) for table in tables}
            self.global_filters[name] = {
                fname: (filter_spec, schemas)
                for fname, filter_spec in filter_specs.items()
            }

    def load_pipelines(self, **kwargs):
        from .pipeline import Pipeline
        pipelines = self.pipelines
        for name, pipeline_spec in self.spec.get('pipelines', {}).items():
            pipelines[name] = Pipeline.from_spec(
                dict(pipeline_spec, name=name, **kwargs)
            )
        return pipelines

    def load_source(self, name: str, source_spec: Dict[str, Any]):
        from .filters.base import Filter
        from .sources.base import Source
        source_spec = dict(source_spec)
        filter_specs = source_spec.pop('filters', None)
        if self.loading_msg:
            self.loading_msg.object = f'Loading {name} source...'
        if name in self.sources:
            source = self.sources[name]
        else:
            source_spec['name'] = name
            source = Source.from_spec(source_spec)
        self.sources[name] = source
        if not filter_specs:
            return source
        filters: Dict[str, Filter] = dict(self.filters.get(name, {}))
        self.filters[name] = filters
        tables = set(filter_spec.get('table') for filter_spec in filter_specs.values())
        if None in tables:
            schemas = source.get_schema()
        else:
            schemas = {table: source.get_schema(table) for table in tables}
        for fname, filter_spec in filter_specs.items():
            if fname not in filters:
                filters[fname] = Filter.from_spec(filter_spec, schemas)
        return source

    def resolve_views(self):
        from .views import View
        exts = []
        for layout in self.spec.get('layouts', []):
            views = layout.get('views', [])
            if isinstance(views, dict):
                views = list(views.values())
            for view in views:
                view_type = View._get_type(view.get('type', None))
                if view_type and view_type._extension:
                    exts.append(view_type._extension)
        for ext in exts:
            __import__(pn.extension._imports[ext])

    def _resolve_source_ref(self, refs: Tuple[str] | Tuple[str, str] | Tuple[str, str, str]):
        if len(refs) == 3:
            refs = cast(Tuple[str, str, str], refs)
            sourceref, table, field = refs
        elif len(refs) == 2:
            refs = cast(Tuple[str, str], refs)
            sourceref, table = refs
        elif len(refs) == 1:
            refs = cast(Tuple[str], refs)
            (sourceref,) = refs

        from .sources.base import Source
        source = Source.from_spec(sourceref)
        if len(refs) == 1:
            return source
        if len(refs) == 2:
            return source.get(table)
        table_schema = source.get_schema(table)
        if field not in table_schema:
            raise ValueError(f"Field '{field}' was not found in "
                             f"'{sourceref}' table '{table}'.")
        field_schema = table_schema[field]
        if 'enum' not in field_schema:
            raise ValueError(f"Field '{field}' schema does not "
                             "declare an enum.")
        return field_schema['enum']

    def resolve_reference(self, reference: str, variables: Variables = None):
        if not is_ref(reference):
            raise ValueError('References should be prefixed by $ symbol.')
        from .variables import Variable
        refs = tuple(reference[1:].split('.'))
        if len(refs) > 3:
            raise ValueError(
                f'Cannot resolve reference {reference!r}. Only variable '
                'and source reference are supported and may have at most '
                'three levels.'
            )
        variables = variables or self.variables
        if refs[0] == 'variables':
            value = variables[refs[1]]
            if isinstance(value, Variable):
                value = value.value
            return value
        return self._resolve_source_ref(refs)  # type: ignore

    def reset(self):
        self._filters.clear()
        self._sources.clear()
        self._pipelines.clear()
        self._specs.clear()
        self._spec.clear()
        self._variables.clear()



state = _session_state()



import argparse

from pathlib import Path

from bokeh.command.subcommand import Argument, Subcommand  # type: ignore
from panel.command import Serve

from ..ui.launcher import Launcher
from ..ui.state import state as ui_state
from ..util import resolve_module_reference


class Builder(Subcommand):

    name = 'builder'

    help = "Launch the Lumen Builder UI"

    args = (
        tuple(
            (name, arg) for name, arg in Serve.args
            if name not in ('files', '--args', '--glob')
        ) + (
        ('--launcher', Argument(
            metavar = 'LAUNCHER',
            type    = str,
            help    = "The Launcher plugin to use",
            default = 'lumen.ui.launcher.LocalLauncher'
        )),
        ('--components', Argument(
            metavar = 'COMPONENTS',
            type    = str,
            help    = "Directory containing components"
        ))
    ))

    def __init__(self, parser: argparse.ArgumentParser, serve: Serve) -> None:
        super().__init__(parser)
        self.serve = serve

    def invoke(self, args: argparse.Namespace) -> None:
        # Set panel serve arguments
        args.glob = False
        args.args = []
        args.files = [str(Path(__file__).parent.parent / 'ui')]

        # Set up UI state
        ui_state.launcher = resolve_module_reference(args.launcher, Launcher)
        if args.components:
            ui_state.components = args.components

        self.serve.invoke(args)



import pytest

from lumen.sources.base import Source
from lumen.validation import ValidationError


@pytest.mark.parametrize(
    "spec,context,msg",
    (
        (
            {"type": "file", "tables": {"penguins": "url.csv"}},
            {"sources": {}},
            None,
        ),
        (
            {"type": "file", "tables": {"penguins": "url.csv"}},
            None,
            None,
        ),
        (
            {"tables": {"penguins": "url.csv"}},
            None,
            "Source component specification did not declare a type",
        ),
        (
            {"type": "afile", "tables": {"penguins": "url.csv"}},
            None,
            "Source component specification declared unknown type 'afile'",
        ),
        (
            "src",
            {"sources": {"src": "Test"}},
            None,
        ),
        (
            "no_src",
            {"sources": {"src": "Test"}},
            "Referenced non-existent source 'no_src'",
        ),
    ),
    ids=[
        "correct1",
        "correct2",
        "missing_type",
        "wrong_type",
        "correct_string_spec",
        "wrong_string_spec",
    ],
)
def test_source_Source(spec, context, msg):
    if msg is None:
        if isinstance(spec, str):
            assert Source.validate(spec, context) == spec
        else:
            assert Source.validate(spec.copy(), context) == spec

    else:
        with pytest.raises(ValidationError, match=msg):
            Source.validate(spec, context)



from pathlib import Path

import pandas as pd

from lumen.panel import DownloadButton
from lumen.sources.base import FileSource
from lumen.state import state
from lumen.variables.base import Variables
from lumen.views.base import View, hvPlotView


def test_resolve_module_type():
    assert View._get_type('lumen.views.base.View') is View


def test_view_hvplot_basis(set_root):
    set_root(str(Path(__file__).parent.parent))
    source = FileSource(tables={'test': 'sources/test.csv'})
    view = {
        'type': 'hvplot',
        'table': 'test',
        'x': 'A',
        'y': 'B',
        'kind': 'scatter',
    }

    view = View.from_spec(view, source, [])

    assert isinstance(view, hvPlotView)
    assert view.kind == 'scatter'
    assert view.pipeline.source is source
    assert view.pipeline.table == 'test'
    assert view.pipeline.filters == []
    assert view.pipeline.transforms == []
    assert view.controls == []

    df = view.get_data()

    assert isinstance(df, pd.DataFrame)
    assert df.shape == (5, 4)

    plot = view.get_plot(df)

    assert plot.kdims == ['A']
    assert plot.vdims == ['B']


def test_view_hvplot_variable(set_root):
    set_root(str(Path(__file__).parent.parent))
    source = FileSource(tables={'test': 'sources/test.csv'})
    state._variable = vars = Variables.from_spec({'y': 'B'})
    view = {
        'type': 'hvplot',
        'table': 'test',
        'x': 'A',
        'y': '$variables.y',
        'kind': 'scatter',
    }

    view = View.from_spec(view, source, [])

    assert view.y == 'B'

    df = view.get_data()
    plot = view.get_plot(df)

    assert plot.vdims == ['B']

    vars._vars['y'].value = 'C'

    plot = view.get_plot(df)

    assert plot.vdims == ['C']


def test_view_hvplot_download(set_root):
    set_root(str(Path(__file__).parent.parent))
    source = FileSource(tables={'test': 'sources/test.csv'})
    view = {
        'type': 'hvplot',
        'table': 'test',
        'x': 'A',
        'y': 'B',
        'kind': 'scatter',
        'download': 'csv'
    }

    view = View.from_spec(view, source)

    button = view.panel[0]
    assert isinstance(button, DownloadButton)

    button._on_click()
    assert button.data.startswith('data:text/plain;charset=UTF-8;base64')

def test_view_to_spec(set_root):
    set_root(str(Path(__file__).parent.parent))
    source = FileSource(tables={'test': 'sources/test.csv'})
    view = {
        'type': 'table',
        'table': 'test',
    }

    view = View.from_spec(view, source, [])
    assert view.to_spec() == {
        'pipeline': {
            'source': {
                'tables': {'test': 'sources/test.csv'},
                'type': 'file'
            },
            'table': 'test'
        },
        'type': 'table',
    }

def test_view_to_spec_with_context(set_root):
    set_root(str(Path(__file__).parent.parent))
    source = FileSource(tables={'test': 'sources/test.csv'})
    view = {
        'type': 'table',
        'table': 'test',
    }

    view = View.from_spec(view, source, [])
    context = {}
    spec = view.to_spec(context=context)
    assert context == {
        'sources': {
            source.name: {
                'tables': {'test': 'sources/test.csv'},
                'type': 'file'
            },
        },
        'pipelines': {
            view.pipeline.name: {
                'source': source.name,
                'table': 'test'
            }
        },
    }
    assert spec == {
        'pipeline': view.pipeline.name,
        'type': 'table'
    }

def test_view_to_spec_with_context_existing_context(set_root):
    set_root(str(Path(__file__).parent.parent))
    source = FileSource(tables={'test': 'sources/test.csv'})
    view = {
        'type': 'table',
        'table': 'test',
    }

    view = View.from_spec(view, source, [])
    context = {}
    view.pipeline.to_spec(context=context)
    assert context == {
        'sources': {
            source.name: {
                'tables': {'test': 'sources/test.csv'},
                'type': 'file'
            },
        },
        'pipelines': {
            view.pipeline.name: {
                'source': source.name,
                'table': 'test'
            }
        },
    }
    spec = view.to_spec(context=context)
    assert spec == {
        'pipeline': view.pipeline.name,
        'type': 'table'
    }

def test_hvplot_view_to_spec(set_root):
    set_root(str(Path(__file__).parent.parent))
    source = FileSource(tables={'test': 'sources/test.csv'})
    view = {
        'type': 'hvplot',
        'table': 'test',
        'x': 'A',
        'y': 'B',
        'alpha': 0.3
    }
    view = View.from_spec(view, source, [])
    assert view.to_spec() == {
        'pipeline': {
            'source': {
                'tables': {'test': 'sources/test.csv'},
                'type': 'file'
            },
            'table': 'test'
        },
        'type': 'hvplot',
        'x': 'A',
        'y': 'B',
        'alpha': 0.3
    }
